#!/usr/bin/env python
# vim: set fileencoding=utf-8
# vim: ts=4:sw=4:et:ai:sts=4

# passfd.py: Python library to pass file descriptors across UNIX domain sockets.
'''This simple extension provides two functions to pass and receive file
descriptors across UNIX domain sockets, using the BSD-4.3+ sendmsg() and
recvmsg() interfaces.

Direct bindings to sendmsg and recvmsg are not provided, as the API does
not map nicely into Python.

Please note that this only supports BSD-4.3+ style file descriptor
passing, and was only tested on Linux. Patches are welcomed!

For more information, see one of the R. Stevens' books:
 - Richard Stevens: Unix Network Programming, Prentice Hall, 1990;
   chapter 6.10

 - Richard Stevens: Advanced Programming in the UNIX Environment,
   Addison-Wesley, 1993; chapter 15.3
'''

#
# Please note that this only supports BSD-4.3+ style file descriptor passing,
# and was only tested on Linux. Patches are welcomed!
#
# Copyright © 2010 Martín Ferrari <martin.ferrari@gmail.com>
#
# Inspired by Socket::PassAccessRights, which is:
#   Copyright (c) 2000 Sampo Kellomaki <sampo@iki.fi>
#
# For more information, see one of the R. Stevens' books:
# - Richard Stevens: Unix Network Programming, Prentice Hall, 1990;
#   chapter 6.10
# 
# - Richard Stevens: Advanced Programming in the UNIX Environment,
#   Addison-Wesley, 1993; chapter 15.3
#
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 2 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#

import os, socket

def __check_socket(sock):
    if hasattr(sock, 'family') and sock.family != socket.AF_UNIX:
        raise ValueError("Only AF_UNIX sockets are allowed")

    if hasattr(sock, 'fileno'):
        sock = sock.fileno()

    if not isinstance(sock, int):
        raise TypeError("An socket object or file descriptor was expected")

    return sock

def __check_fd(fd):
    try:
        fd = fd.fileno()
    except AttributeError:
        pass
    if not isinstance(fd, int):
        raise TypeError("An file object or file descriptor was expected")

    return fd

def sendfd(sock, fd, message = "NONE"):
    """Sends a message and piggybacks a file descriptor through a Unix
    domain socket.

    Note that the file descriptor cannot be sent by itself, at least
    one byte of payload needs to be sent also.

    Parameters:
     sock:    socket object or file descriptor for an AF_UNIX socket
     fd:      file object or file descriptor to pass
     message: message to send

    Return value:
    On success, sendfd returns the number of bytes sent, not including
    the file descriptor nor the control data.  If there was no message
    to send, 0 is returned."""
    
    import _passfd
    return _passfd.sendfd(__check_socket(sock), __check_fd(fd), message)

def recvfd(sock, msg_buf = 4096):
    """Receive a message and a file descriptor from a Unix domain socket.
    
    Parameters:
     sock:       file descriptor or socket object for an AF_UNIX socket
     buffersize: maximum message size to receive

    Return value:
    On success, recvfd returns a tuple containing the received
    file descriptor and message. If recvmsg fails, an OSError exception
    is raised. If the received data does not carry exactly one file
    descriptor, or if the received file descriptor is not valid,
    RuntimeError is raised."""

    import _passfd
    (ret, msg) = _passfd.recvfd(__check_socket(sock), msg_buf)

    # -1 should raise OSError
    if ret == -2:
        raise RuntimeError("The message received did not contain exactly one" +
                " file descriptor")
    if ret == -3:
        raise RuntimeError("The received file descriptor is not valid")
    assert ret >= 0

    return (ret, msg)



#!/usr/bin/env python
# vim: set fileencoding=utf-8
# vim: ts=4:sw=4:et:ai:sts=4
from distutils.core import setup, Extension
import sys
sys.path.append("src")
import passfd

module1 = Extension('_passfd', sources = ['src/passfd.c'])

setup(
        name        = 'python-passfd',
        version     = '0.2',
        description = 'Python functions to pass file descriptors across ' +
        'UNIX domain sockets',
        long_description = passfd.__doc__,
        author      = 'Martin Ferrari',
        author_email = 'martin.ferrari@gmail.com',
        url         = 'http://code.google.com/p/python-passfd/',
        license     = 'GPLv2',
        platforms   = 'Linux',
        package_dir = {'': 'src'},
        ext_modules = [module1],
        py_modules  = ['passfd'])



#!/usr/bin/env python
# vim: set fileencoding=utf-8
# vim: ts=4:sw=4:et:ai:sts=4

# Copyright © 2010 Martín Ferrari <martin.ferrari@gmail.com>
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 2 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#

import os, unittest, socket, sys, threading
from passfd import sendfd, recvfd
import _passfd

class TestPassfd(unittest.TestCase):
    def readfd_test(self, fd):
        s = os.read(fd, 512)
        self.assertEquals(s, "\0" * 512)

    def vrfy_recv(self, tuple, msg):
        self.readfd_test(tuple[0])
        self.assertEquals(tuple[1], msg)

    def parent_tests(self, s, dgram = False):
        # First message is not even sent
        s.send("1")
        self.vrfy_recv(recvfd(s), "a")
        s.send("2")
        self.vrfy_recv(recvfd(s), "\0")
        s.send("3")
        self.vrfy_recv(recvfd(s), "foobar")
        s.send("4")
        self.vrfy_recv(recvfd(s, msg_buf = 11), "long string") # is long
        if not dgram:
            self.assertEquals(s.recv(8), " is long") # re-sync
        s.send("5")
        self.assertEquals(s.recv(100), "foobar")
        s.send("6")
        self.assertRaises(RuntimeError, recvfd, s) # No fd received
        #
        s.send("7")
        f, m = recvfd(s)
        self.assertRaises(OSError, os.fdopen, f, "w")
        s.send("8")
        (f, msg) = recvfd(s)
        self.assertEquals(msg, "writing")
        os.write(f, "foo")
        s.send("9")

    def child_tests(self, s, dgram = False):
        f = file("/dev/zero")
        assert sendfd(s, f, "") == 0
        s.recv(1)
        assert sendfd(s, f, "a") == 1
        s.recv(1)
        assert sendfd(s, f, "\0") == 1
        s.recv(1)
        assert sendfd(s, f, "foobar") == 6
        s.recv(1)
        assert sendfd(s, f, "long string is long") == 19
        # The other side will recv() instead of recvmsg(), this fd would be
        # lost. I couldn't find any specification on this semantic
        s.recv(1)
        assert sendfd(s, f, "foobar") == 6
        s.recv(1)
        assert s.send("barbaz") == 6
        # Try to write!
        s.recv(1)
        assert sendfd(s, f, "writing") == 7
        s.recv(1)
        f = file("/dev/null", "w")
        assert sendfd(s, f, "writing") == 7
        s.recv(1)

    def test_sanity_checks(self):
        self.assertRaises(TypeError, recvfd, "foo")
        s = socket.socket(socket.AF_INET)
        self.assertRaises(ValueError, recvfd, s)

        (s0, s1) = socket.socketpair(socket.AF_UNIX, socket.SOCK_STREAM, 0)
        f = file("/dev/zero")
        sendfd(s0, f)
        recvfd(s1)

        # Using integers
        sendfd(s0.fileno(), f.fileno())
        recvfd(s1.fileno())

        self.assertRaises(TypeError, sendfd, s0, "foo")
        # Assuming fd 255 is not valid
        self.assertRaises(OSError, sendfd, s0, 255)

    def test_passfd_stream(self):
        (s0, s1) = socket.socketpair(socket.AF_UNIX, socket.SOCK_STREAM, 0)
        pid = os.fork()
        if pid == 0:
            s0.close()
            self.child_tests(s1)
            s1.close()
            os._exit(0)

        s1.close()
        self.parent_tests(s0)
        s0.close()

        self.assertEquals(os.waitpid(pid, 0)[1], 0)

    def test_passfd_dgram(self):
        (s0, s1) = socket.socketpair(socket.AF_UNIX, socket.SOCK_DGRAM, 0)
        pid = os.fork()
        if pid == 0:
            s0.close()
            self.child_tests(s1, dgram = True)
            s1.close()
            os._exit(0)

        s1.close()
        self.parent_tests(s0, dgram = True)
        s0.close()

        self.assertEquals(os.waitpid(pid, 0)[1], 0)

    def test_threading(self):
        # Check that the GIL is correctly released before blocking
        (s0, s1) = socket.socketpair(socket.AF_UNIX, socket.SOCK_STREAM, 0)

        def run_server(s):
            self.child_tests(s)
            s.close()

        t = threading.Thread(target = run_server, args = (s1,))
        t.start()
        self.parent_tests(s0)
        s0.close()
        t.join()

if __name__ == '__main__':
    unittest.main()

